#include "ktnefparser.h"
#include "ktnefattach.h"

#include <iostream.h>

#define TNEF_SIGNATURE   0x223e9f78
#define LVL_MESSAGE      0x01
#define LVL_ATTACHMENT   0x02
#define attATTACHDATA	0x800f	/* Attachment Data */
#define attATTACHTITLE	0x8010	/* Attachment File Name */
#define attATTACHMENT	0x9005	/* Attachment properties (???) */

KTNEFParser::KTNEFParser()
{
	attachments_.setAutoDelete(true);
	defaultdir_ = "/tmp/";
	current_ = 0;
}

KTNEFParser::~KTNEFParser()
{
	if (file_.isOpen()) file_.close();
}

bool KTNEFParser::decodeMessage()
{
	Q_UINT32	i;
	Q_UINT16	u;

	// read (type+name)
	stream_ >> i;
	// read data length
	stream_ >> i;
	// skip data
	if ( !file_.at( file_.at() + i ) )
		return false;
	// get checksum
	stream_ >> u;
	return true;
}

bool KTNEFParser::decodeAttachment()
{
	Q_UINT32	d;
	Q_UINT16	u;
	char		*buf;
	QByteArray	*array;

	stream_ >> d;		// d <- attribute type & name
	u = ((d << 16) >> 16);	// u <- attribute name
	stream_ >> d;		// d <- data length
	switch (u)
	{
	   case attATTACHTITLE:
		checkCurrent(KTNEFAttach::TitleParsed);
		buf = new char[d+1];
		stream_.readRawBytes(buf,d);
		buf[d] = 0;
		current_->setName(QString(buf));
		delete [] buf;
		break;
	   case attATTACHDATA:
		checkCurrent(KTNEFAttach::DataParsed);
		current_->setSize(d);
		current_->setOffset(file_.at());
		file_.at(file_.at()+d);
		break;
	   case attATTACHMENT:	// try to get attachment info
		checkCurrent(KTNEFAttach::InfoParsed);
		buf = new char[d];
		stream_.readRawBytes(buf,d);
		array = new QByteArray;
		array->assign(buf,d);
		decodeAttachmentInfo(array);
		delete array;	// will also delete the "buf" structure
		break;
	   default:
		// simply skip data and checksum
		file_.at( file_.at() + d );
		break;
	}
	stream_ >> u;	// u <- checksum

	return true;
}

void KTNEFParser::setDefaultExtractDir(const QString& dirname)
{
	defaultdir_ = dirname;
}

bool KTNEFParser::parseFile(const QString& filename)
{
	Q_UINT16	u;
	Q_UINT32	d;
	Q_UINT8		c;

	if (file_.isOpen()) file_.close();
	attachments_.clear();
	if (current_)
	{
		delete current_;
		current_ = 0;
	}

	file_.setName(filename);
	if (!file_.exists() || !file_.open(IO_ReadOnly)) return false;
	stream_.setDevice(&file_);
	stream_.setByteOrder(QDataStream::LittleEndian);
	stream_ >> d;
	if (d == TNEF_SIGNATURE)
	{
		stream_ >> u;
		while (!stream_.eof())
		{
			stream_ >> c;
			switch (c)
			{
			   case LVL_MESSAGE:
				if (!decodeMessage()) goto end;
				break;
			   case LVL_ATTACHMENT:
				if (!decodeAttachment()) goto end;
				break;
			   default:
				goto end;
			}
		}
		if (current_)
		{
			checkCurrent(KTNEFAttach::DataParsed);	// this line has the effect to append the
								// attachment, if it has data. If not it does
								// nothing, and the attachment will be discarded
			delete current_;
			current_ = 0;
		}
		return true;
	}
	else
	{
end:		file_.close();
		return false;
	}
}

KTNEFAttach* KTNEFParser::getAttachment(const QString& filename)
{
	QListIterator<KTNEFAttach>	it(attachments_);
	for (;it.current();++it)
		if (it.current()->name() == filename) return it.current();
	return 0;
}

bool KTNEFParser::extractFile(const QString& filename)
{
	KTNEFAttach	*att = getAttachment(filename);
	if (!att) return false;
	return extractAttachmentTo(att, defaultdir_);
}

bool KTNEFParser::extractAttachmentTo(KTNEFAttach *att, const QString& dirname)
{
	QString	filename = dirname + "/" + att->name();
	if (!file_.isOpen()) return false;
	if (!file_.at(att->offset())) return false;
	QFile	outfile(filename);
	if (!outfile.open(IO_WriteOnly)) return false;

	Q_UINT32	len = att->size(), sz(16384);
	int		n(0);
	char		*buf = new char[sz];
	bool		done(false);
	while (len > 0)
	{
		n = file_.readBlock(buf,QMIN(sz,len));
		if (n < 0) goto end;
		len -= n;
		if (outfile.writeBlock(buf,n) != n) goto end;
	}
	done = true;

end:	outfile.close();
	return done;
}

bool KTNEFParser::extractAll()
{
	QListIterator<KTNEFAttach>	it(attachments_);
	for (;it.current();++it)
		if (!extractAttachmentTo(it.current(),defaultdir_)) return false;
	return true;
}

bool KTNEFParser::extractFileTo(const QString& filename, const QString& dirname)
{
	KTNEFAttach	*att = getAttachment(filename);
	if (!att) return false;
	return extractAttachmentTo(att, dirname);
}

bool KTNEFParser::openFile(const QString& filename)
{
	filename_ = filename;
	return parseFile(filename);
}

void KTNEFParser::checkCurrent(int state)
{
	if (!current_) current_ = new KTNEFAttach();
	else
	{
		if (current_->checkState(state))
		{
			if (current_->offset() >= 0 )
			{
				if (current_->name().isEmpty()) current_->setName("Unnamed");
				attachments_.append(current_);
				current_ = 0;
			}
			else
			{ // invalid attachment, skip it
				delete current_;
				current_ = 0;
			}
			current_ = new KTNEFAttach();
		}
	}
}

//----------------------------------------------------------------------------------------

// supported MAPI types
#define	MAPI_TYPE_NONE		0x0000
#define	MAPI_TYPE_UINT16	0x0002
#define	MAPI_TYPE_ULONG		0x0003
#define	MAPI_TYPE_FLOAT		0x0004
#define	MAPI_TYPE_DOUBLE	0x0005
#define	MAPI_TYPE_TIME		0x0040
#define	MAPI_TYPE_STRING8	0x001e
#define	MAPI_TYPE_USTRING	0x001f
#define	MAPI_TYPE_BINARY	0x0102

// supported MAPI tags
#define	MAPI_TAG_INDEX		0x0e21
#define	MAPI_TAG_SIZE		0x0e20
#define	MAPI_TAG_FILENAME	0x3707
#define	MAPI_TAG_DISPLAYNAME	0x3001
#define	MAPI_TAG_MIMETAG	0x370e
#define	MAPI_TAG_EXTENSION	0x3703

#define	FREE(buf) if(buf) { delete [] buf; buf = 0; }
#define	ALIGN(n,b) (n) = ((((n)/(b)) + 1) * (b))

typedef struct {
	int	type;
	int	len;
	union {
		Q_UINT16	int16;
		Q_UINT32	int32;
		float		real;
		double		doublereal;
		char*		buffer;
	} value;
} MAPI_value;

void clearMAPIValue(MAPI_value& mapi)
{
	if (mapi.type == MAPI_TYPE_STRING8 || mapi.type == MAPI_TYPE_USTRING || mapi.type == MAPI_TYPE_BINARY)
		FREE(mapi.value.buffer);
	mapi.value.doublereal = 0;
}

Q_UINT16 readMAPIValue(QDataStream& stream, MAPI_value& mapi)
{
	Q_UINT16	mapitag;
	Q_UINT32	d;

	clearMAPIValue(mapi);
	stream >> d;
	mapi.type =  (d & 0x0000FFFF);
	mapitag = ((d & 0xFFFF0000) >> 16);
	switch(mapi.type)
	{
	   case MAPI_TYPE_UINT16:
		stream >> mapi.value.int16;
		break;
	   case MAPI_TYPE_ULONG:
		stream >> mapi.value.int32;
		break;
	   case MAPI_TYPE_FLOAT:
		stream >> mapi.value.real;
		break;
	   case MAPI_TYPE_DOUBLE:
		stream >> mapi.value.doublereal;
		break;
	   case MAPI_TYPE_TIME:
		stream >> d >> d;
		break;
	   case MAPI_TYPE_STRING8:
		stream >> d;
		for (uint i=0;i<d;i++)
		{
			clearMAPIValue(mapi);
			stream >> mapi.len;
			if (mapi.len > 0)
			{
				ALIGN(mapi.len, 4);
				mapi.value.buffer = new char[mapi.len+1];
				stream.readRawBytes(mapi.value.buffer, mapi.len);
				mapi.value.buffer[mapi.len] = 0;
			}
		}
		break;
	   case MAPI_TYPE_USTRING:
		mapi.type = MAPI_TYPE_NONE;
		break;
	   case MAPI_TYPE_BINARY:
		stream >> d;
		for (uint i=0;i<d;i++)
		{
			clearMAPIValue(mapi);
			stream >> mapi.len;
			if (mapi.len > 0)
			{
				ALIGN(mapi.len, 4);
				mapi.value.buffer = new char[mapi.len];
				stream.readRawBytes(mapi.value.buffer, mapi.len);
			}
		}
		break;
	   default:
		mapi.type = MAPI_TYPE_NONE;
		break;
	}
	return mapitag;
}

bool KTNEFParser::decodeAttachmentInfo(QByteArray *array)
{
	QDataStream	infostream_(*array, IO_ReadOnly);
	Q_UINT16	mapitag;
	Q_UINT32	n;
	MAPI_value	mapi;

	// some initializations
	mapi.type = MAPI_TYPE_NONE;
	mapi.value.buffer = 0;

	infostream_.setByteOrder(QDataStream::LittleEndian);
	// get number of properties
	infostream_ >> n;
	for (uint i=0;i<n;i++)
	{
		if (infostream_.eof())
		{
			clearMAPIValue(mapi);
			return false;
		}
		mapitag = readMAPIValue(infostream_, mapi);
		if (mapi.type == MAPI_TYPE_NONE || !current_)
		{
			clearMAPIValue(mapi);
			return false;
		}
		switch (mapitag)
		{
		   case MAPI_TAG_INDEX:
			current_->setIndex(mapi.value.int32);
			break;
		   case MAPI_TAG_SIZE:
			current_->setDisplaySize(mapi.value.int32);
			break;
		   case MAPI_TAG_FILENAME:
			current_->setFileName(QString(mapi.value.buffer));
			break;
		   case MAPI_TAG_DISPLAYNAME:
			current_->setDisplayName(QString(mapi.value.buffer));
			break;
		   case MAPI_TAG_MIMETAG:
			current_->setMimeTag(QString(mapi.value.buffer));
			break;
		   case MAPI_TAG_EXTENSION:
			current_->setExtension(QString(mapi.value.buffer));
			break;
		}
	}
	return true;
}
/*
bool KTNEFParser::decodeAttachmentInfo(QByteArray *array)
{
	QDataStream	infostream_(*array, IO_ReadOnly);
	Q_UINT16	u1, u2;
	Q_UINT32	d1, d2;

	float		fl;
	double		dbl;
	char*		buf(0);
	int		n;

	infostream_.setByteOrder(QDataStream::LittleEndian);
	// skip first integer, then loop...
	infostream_ >> d1;
	while (!infostream_.eof())
	{
		FREE(buf);
		infostream_ >> d1;	// d1 <- type + tag
		u1 = ((d1 << 16) >> 16);	// u1 <- type
		u2 = (d1 >> 16);		// u2 <- tag

		// first read the data
		switch (u1)
		{
		   case MAPI_TYPE_UINT16:
			infostream_ >> u1;
			n = u1;
			break;
		   case MAPI_TYPE_ULONG:
			infostream_ >> d1;
			n = d1;
			break;
		   case MAPI_TYPE_FLOAT:
			infostream_ >> fl;
			break;
		   case MAPI_TYPE_DOUBLE:
			infostream_ >> dbl;
			break;
		   case MAPI_TYPE_TIME:
			buf = new char[8];
			infostream_.readRawBytes(buf, 8);
			break;
		   case MAPI_TYPE_STRING8:
			infostream_ >> d1;	// number of strings
			for (uint i=0;i<d1;i++)
			{
				FREE(buf);
				infostream_ >> n;	// length of string
				if (n != 0)
				{
					ALIGN(n,4);
					buf = new char[n+1];
					infostream_.readRawBytes(buf,n);
					buf[n] = 0;
				}
			}
			break;
		   case MAPI_TYPE_USTRING:	// not really supported
			FREE(buf);
			return false;
		   case MAPI_TYPE_BINARY:
			infostream_ >> d1;	// number of binary arrays
			for (uint i=0;i<d1;i++)
			{
				FREE(buf);
				infostream_ >> n;	// array size
				if (n != 0)
				{
					ALIGN(n,4);
					buf = new char[n];
					infostream_.readRawBytes(buf,n);
				}
			}
			break;
		   default:	// unsupported
			FREE(buf);
			return false;
		}

		// then interpret the data
		if (!current_)
		{
			FREE(buf);
			return false;
		}
		switch (u2)
		{
		   case MAPI_TAG_INDEX:
			current_->setIndex(n);
			break;
		   case MAPI_TAG_SIZE:
			current_->setDisplaySize(n);
			break;
		   case MAPI_TAG_FILENAME:
			current_->setFileName(QString(buf));
			break;
		   case MAPI_TAG_DISPLAYNAME:
			current_->setDisplayName(QString(buf));
			break;
		   case MAPI_TAG_MIMETAG:
			current_->setMimeTag(QString(buf));
			break;
		   case MAPI_TAG_EXTENSION:
			current_->setExtension(QString(buf));
			break;
		}
	}
	return true;
}
*/
